local binaryWriter = require("mods").requireFromPlugin("binary_writer")
local binaryReader = require("mods").requireFromPlugin("binary_reader")

-- Allows tables to use custom _type value
local function typeof(value)
    local valueType = type(value)

    if valueType == "table" then
        return value._type or "table"

    else
        return valueType
    end
end

local serialize = {}

-- Default type numbers
-- 0 - Nil
-- 1 - Boolean
-- 2 - Int8
-- 3 - Int16
-- 4 - Int32
-- 5 - Float32
-- 6 - String


-- Assume anything that doesn't fit is Int64
local integerBits = {
    {2, 0, 255, "writeByte"},
    {3, -32768, 32768, "writeSignedShort"},
    {4, -2147483648, 2147483647, "writeSignedLong"}
}

function serialize.serializeNil(writer, value)
    writer:writeVariableLength(0)
end

function serialize.unserializeNil(value)

end

function serialize.serializeBoolean(writer, value)
    writer:writeVariableLength(1)
    writer:writeBool(value)
end

function serialize.unserializeBoolean(reader)
    return reader:readBool()
end

function serialize.serializeNumber(writer, value)
    local isInteger = value == math.floor(value)

    if isInteger then
        for _, bits in ipairs(integerBits) do
            local typeNumber, min, max, func = bits[1], bits[2], bits[3], bits[4]

            if value >= min and value <= max then
                writer:writeVariableLength(typeNumber)
                writer[func](writer, value)

                return
            end
        end

    else
        writer:writeVariableLength(5)
        writer:writeFloat(value)
    end
end

function serialize.unserializeByte(reader)
    return reader:readByte()
end

function serialize.unserializeSignedShort(reader)
    return reader:readSignedShort()
end

function serialize.unserializeSignedLong(reader)
    return reader:readSignedLong()
end

function serialize.unserializeFloat(reader)
    return reader:readFloat()
end

function serialize.serializeString(writer, value)
    writer:writeVariableLength(6)
    writer:writeString(value)
end

function serialize.unserializeString(reader)
    return reader:readString()
end

function serialize.serializeTable(writer, value, funcs)
    local numericalKeys = 0
    local allKeys = 0

    for k, v in pairs(value) do
        allKeys = allKeys + 1
    end

    for _, v in ipairs(value) do
        numericalKeys = numericalKeys + 1
    end

    if allKeys == numericalKeys then
        writer:writeVariableLength(7)
        writer:writeVariableLength(numericalKeys)

        for _, v in ipairs(value) do
            serialize.serialize(v, funcs, writer)
        end

    else
        writer:writeVariableLength(8)
        writer:writeVariableLength(allKeys)

        for k, v in pairs(value) do
            serialize.serialize(k, funcs, writer)
            serialize.serialize(v, funcs, writer)
        end
    end
end

function serialize.unserializeNumericalTable(reader, funcs)
    local res = {}
    local length = reader:readVariableLength()

    for i = 1, length do
        local v = serialize.unserialize(nil, funcs, reader)

        res[i] = v
    end

    return res
end

function serialize.unserializeNonNumericalTable(reader, funcs)
    local res = {}
    local length = reader:readVariableLength()

    for i = 1, length do
        local k = serialize.unserialize(nil, funcs, reader)
        local v = serialize.unserialize(nil, funcs, reader)

        res[k] = v
    end

    return res
end


serialize._defaultSerialization = {
    ["nil"]     = serialize.serializeNil,
    ["boolean"] = serialize.serializeBoolean,
    ["number"]  = serialize.serializeNumber,
    ["string"]  = serialize.serializeString,
    ["table"]   = serialize.serializeTable,
}

serialize._defaultDeserialization = {
    [0] = serialize.unserializeNil,
    [1] = serialize.unserializeBoolean,
    [2] = serialize.unserializeByte,
    [3] = serialize.unserializeSignedShort,
    [4] = serialize.unserializeSignedLong,
    [5] = serialize.unserializeFloat,
    [6] = serialize.unserializeString,
    [7] = serialize.unserializeNumericalTable,
    [8] = serialize.unserializeNonNumericalTable,
}


function serialize.serialize(value, funcs, writer)
    writer = writer or binaryWriter()

    local valueType = typeof(value)

    if funcs and funcs[valueType] then
        funcs[valueType](writer, value, funcs)

    else
        serialize._defaultSerialization[valueType](writer, value, funcs)
    end

    return writer:getString()
end

function serialize.unserialize(value, funcs, reader)
    reader = reader or binaryReader(value)

    local readType = reader:readVariableLength()

    if funcs and funcs[readType] then
        return funcs[readType](reader, funcs)

    else
        return serialize._defaultDeserialization[readType](reader, funcs)
    end
end

return serialize