-- Pure Lua binary file implementation

local floor = math.floor

local binfile = {}

function binfile.twosCompliment(n, power)
    if n >= 2^(power - 1) then
        return n - 2^power

    else
        return n
    end
end

function binfile.readByte(fh)
    return string.byte(fh:read(1))
end

function binfile.writeByte(fh, n)
    fh:write(string.char(n))
end

function binfile.writeBytes(fh, ...)
    fh:write(string.char(...))
end

function binfile.readShort(fh)
    local b1, b2 = string.byte(fh:read(2), 1, 2)

    return b1 + b2 * 256
end

function binfile.writeShort(fh, n)
    binfile.writeBytes(fh, n % 256, floor(n / 256) % 256)
end

function binfile.readLong(fh)
    local b1, b2, b3, b4 = string.byte(fh:read(4), 1, 4)

    return b1 + b2 * 256 + b3 * 65536 + b4 * 16777216
end

function binfile.writeLong(fh, n)
    binfile.writeBytes(fh, n % 256, floor(n / 256) % 256, floor(n / 65536) % 256, floor(n / 16777216) % 256)
end

function binfile.readBool(fh)
    return binfile.readByte(fh) ~= 0
end

function binfile.writeBool(fh, b)
    binfile.writeByte(fh, b and 1 or 0)
end

function binfile.readSignedShort(fh)
    return binfile.twosCompliment(binfile.readShort(fh), 16)
end

function binfile.writeSignedShort(fh, n)
    binfile.writeShort(fh, binfile.twosCompliment(n, 16))
end

function binfile.readSignedLong(fh)
    return binfile.twosCompliment(binfile.readLong(fh), 32)
end

function binfile.writeSignedLong(fh, n)
    binfile.writeLong(fh, binfile.twosCompliment(n, 32))
end

function binfile.readFloat(fh)
    local b4, b3, b2, b1 = string.byte(fh:read(4), 1, 4)

    local exponent = (b1 % 128) * 2 + floor(b2 / 128)

    if exponent == 0 then
        return 0.0
    end

    local sign = (b1 > 127) and -1 or 1
    local mantissa = ((b2 % 128) * 256 + b3) * 256 + b4
    mantissa = (math.ldexp(mantissa, -23) + 1) * sign

    return math.ldexp(mantissa, exponent - 127)
end

function binfile.writeFloat(fh, n)
    local b1, b2, b3, b4

    local val = n
    local sign = 0

    if val < 0 then
        sign = 1
        val = -val
    end

    local mantissa, exponent = math.frexp(val)

    if val == 0 then
        mantissa = 0
        exponent = 0

    else
        mantissa = (mantissa * 2 - 1) * math.ldexp(0.5, 24)
        exponent = exponent + 126
    end

    b1 = floor(mantissa) % 256
    val = floor(mantissa / 256)
    b2 = floor(val) % 256
    val = floor(val / 256)

    b3 = floor(exponent * 128 + val) % 256
    val = floor((exponent * 128 + val) / 256)
    b4 = floor(sign * 128 + val) % 256

    binfile.writeBytes(fh, b1, b2, b3, b4)
end

function binfile.readVariableLength(fh)
    local res = 0
    local multiplier = 1

    while true do
        local byte = binfile.readByte(fh)

        if byte < 128 then
            return res + byte * multiplier

        else
            res = res + (byte - 128) * multiplier
        end

        multiplier = multiplier * 128
    end
end

function binfile.getVariableLength(fh, length)
    local res = {}

    while length > 127 do
        table.insert(res, length % 128 + 128)

        length = floor(length / 128)
    end

    table.insert(res, length)

    return res
end

function binfile.writeVariableLength(fh, length)
    while length > 127 do
        binfile.writeByte(fh, length % 128 + 128)

        length = floor(length / 128)
    end

    binfile.writeByte(fh, length)
end

function binfile.readString(fh)
    local length = binfile.readVariableLength(fh)
    local res = fh:read(length)

    return res
end

function binfile.writeString(fh, value)
    binfile.writeVariableLength(fh, #value)
    fh:write(value)
end

function binfile.readRunLengthEncoded(fh)
    local bytes = binfile.readShort(fh)
    local res = {}

    for i = 1, bytes, 2 do
        local times = binfile.readByte(fh)
        local char = fh:read(1)

        table.insert(res, char:rep(times))
    end

    return table.concat(res)
end

function binfile.writeByteArray(fh, t)
    fh:write(string.char(unpack(t)))
end

return binfile