local utils = require("mods").requireFromPlugin("utils")

local excitingBrickGameMovement = {}

local clearedLinesName = {
    "Single",
    "Double",
    "Triple",
    "Exciting!"
}

local clearedLinesNameTSpin = {
    [0] = "T-Spin",
    "T-Spin Single",
    "T-Spin Double",
    "T-Spin Triple"
}

local clearedLinesNameTSpinMini = {
    [0] = "T-Spin Mini",
    "T-Spin Mini Single",
}

local allClearName = "All Clear"

-- Name of the current clear
-- For example: Exciting!, Single, T-Spin Double, All Clear
function excitingBrickGameMovement.lineClearName(clearedLines, tspin, allClear)
    if allClear then
        return allClearName

    elseif tspin == "proper" then
        return clearedLinesNameTSpin[clearedLines]

    elseif tspin == "mini" then
        return clearedLinesNameTSpinMini[clearedLines]

    else
        return clearedLinesName[clearedLines]
    end
end

-- Returns whether a non air part of the excitingBrickShape is colliding with the playfield
function excitingBrickGameMovement.exciting_brick_shapeColliding(excitingBrickShape, playfield)
    if excitingBrickShape then
        local excitingBrickShapeMatrix = excitingBrickShape:getMatrix()
        local playfieldMatrix = playfield._playfieldMatrix
        local width, height = excitingBrickShapeMatrix:size()

        for x = 1, width do
            for y = 1, height do
                local playfieldMino = playfieldMatrix:get(excitingBrickShape.x - excitingBrickShape.centerX + x, excitingBrickShape.y - excitingBrickShape.centerY + y, "g")
                local mino = excitingBrickShapeMatrix:get(x, y, " ")

                if mino ~= " " and playfieldMino ~= " " then
                    return true
                end
            end
        end

        return false
    end

    return false
end

-- Count the diagonally adjacent tiles for the center of T excitingBrickShape
-- Any tile that isnt air is considered a solid tile
function excitingBrickGameMovement.countTSpinDiagonalCorners(excitingBrickShape, playfield)
    if excitingBrickShape and excitingBrickShape.name == "t" then
        local x, y = excitingBrickShape.x, excitingBrickShape.y
        local matrix = playfield._playfieldMatrix

        return (matrix:get(x - 1, y - 1, "g") ==  " " and 0 or 1) +
            (matrix:get(x + 1, y - 1, "g") ==  " " and 0 or 1) +
            (matrix:get(x - 1, y + 1, "g") ==  " " and 0 or 1) +
            (matrix:get(x + 1, y + 1, "g") ==  " " and 0 or 1)
    end

    return 0
end

-- Count the diagonal corners in the direction of the T excitingBrickShape
-- Any tile that isnt air is considered a solid tile
function excitingBrickGameMovement.countTSpinPointingCorners(excitingBrickShape, playfield)
    if excitingBrickShape and excitingBrickShape.name == "t" then
        local x, y = excitingBrickShape.x, excitingBrickShape.y
        local rotation = excitingBrickShape.rotation
        local matrix = playfield._playfieldMatrix

        -- Top left going clockwise
        local a = (matrix:get(x - 1, y - 1, "g") ==  " " and 0 or 1)
        local b = (matrix:get(x + 1, y - 1, "g") ==  " " and 0 or 1)
        local c = (matrix:get(x - 1, y + 1, "g") ==  " " and 0 or 1)
        local d = (matrix:get(x + 1, y + 1, "g") ==  " " and 0 or 1)

        if rotation == 1 then
            return a + b

        elseif rotation == 2 then
            return b + c

        elseif rotation == 3 then
            return c + d

        elseif rotation == 4 then
            return d + a
        end
    end

    return 0
end

-- Check if the rotation is a T-Spin Mini, T-Spin or just a regular spin
-- Uses implementation for Exciting! Friends, seems to be what Puyo Puyo Exciting! implements
-- TODO - Add option for 3-corner kick/no kick?
function excitingBrickGameMovement.checkTSpinLevel(excitingBrickShape, playfield, offsetX, offsetY, offsetIndex, direction)
    if excitingBrickShape and excitingBrickShape.name == "t" then
        local level = nil

        local rotation = excitingBrickShape.rotation
        local hadKick = offsetX ~= 0 or offsetY ~= 0
        local finalKickOffset = offsetIndex == #excitingBrickShape.offsets[rotation]

        local solidDiagonalCorners = excitingBrickGameMovement.countTSpinDiagonalCorners(excitingBrickShape, playfield)
        local solidPointingCorners = excitingBrickGameMovement.countTSpinPointingCorners(excitingBrickShape, playfield)

        local pointingLeftOrRight = rotation == 2 or rotation == 4

        local potentialProper = not hadKick and solidDiagonalCorners == 3
        local potentialMini = hadKick and solidDiagonalCorners == 3

        -- Normal setup
        if potentialProper or potentialMini then
            if solidPointingCorners == 2 then
                level = "proper"

            elseif solidPointingCorners == 1 then
                level = "mini"
            end
        end

        -- Check if "tripple twist" setup
        if pointingLeftOrRight and finalKickOffset then
            level = "proper"
        end

        return level
    end
end

-- Locks a excitingBrickShape onto the field
function excitingBrickGameMovement.lockDownexcitingBrickShape(excitingBrickShape, playfield)
    if excitingBrickShape then
        local excitingBrickShapeMatrix = excitingBrickShape:getMatrix()
        local playfieldMatrix = playfield._playfieldMatrix
        local width, height = excitingBrickShapeMatrix:size()

        for x = 1, width do
            for y = 1, height do
                local mino = excitingBrickShapeMatrix:get(x, y, " ")

                if mino ~= " " then
                    playfieldMatrix:set(excitingBrickShape.x - excitingBrickShape.centerX + x, excitingBrickShape.y - excitingBrickShape.centerY + y, mino)
                end
            end
        end
    end
end

-- Find lines that need clearing
-- Returns filled lines and whether the playfield would be empty after clearing these (all clear)
function excitingBrickGameMovement.findFilledLines(playfield)
    local solidLines = {}
    local emptyLines = {}
    local partialLines = {}

    local matrix = playfield._playfieldMatrix
    local width, height = matrix:size()

    for row = 1, height do
        local fullLine = true
        local emptyLine = true

        for col = 1, width do
            local targetTile = matrix:get(col, row, "g")

            if targetTile == " " then
                fullLine = false

            else
                emptyLine = false
            end
        end

        if fullLine then
            table.insert(solidLines, row)

        elseif emptyLine then
            table.insert(emptyLines, row)

        else
            table.insert(partialLines, row)
        end
    end

    return solidLines, #partialLines == 0
end

-- Copy field down by offset
function excitingBrickGameMovement.moveFieldDown(playfield, stop, offset)
    local matrix = playfield._playfieldMatrix
    local width, height = matrix:size()

    for row = stop, 1, -1 do
        for col = 1, width do
            matrix:set(col, row, matrix:get(col, row - offset, " "))
        end
    end
end

-- Move whole field up by amount
function excitingBrickGameMovement.moveFieldUp(playfield, amount)
    local matrix = playfield._playfieldMatrix
    local width, height = matrix:size()

    for row = 1, height do
        for col = 1, width do
            matrix:set(col, row, matrix:get(col, row + amount , " "))
        end
    end
end

function excitingBrickGameMovement.setRow(playfield, row, data)
    local matrix = playfield._playfieldMatrix
    local width, height = matrix:size()

    if #data ~= width then
        return false, "Data length must be same as playfield width"
    end

    for x = 1, width do
        matrix:set(x, row, data[x])
    end

    return true
end

-- Generate a line of garbage with a hole at hole
function excitingBrickGameMovement.generateGarbageRow(width, hole)
    local res = {}

    for i = 1, width do
        res[i] = i == hole and " " or "g"
    end

    return res
end

-- Add garbage blocks to the field
-- TODO - Use custom random?
function excitingBrickGameMovement.addGarbageToPlayfield(playfield, lines, mode)
    -- Nothing to do
    if lines <= 0 then
        return
    end

    local matrix = playfield._playfieldMatrix
    local width, height = matrix:size()

    excitingBrickGameMovement.moveFieldUp(playfield, lines)

    if mode == "well" then
        local hole = math.random(1, width)

        for i = 1, lines do
            local garbage = excitingBrickGameMovement.generateGarbageRow(width, hole)

            excitingBrickGameMovement.setRow(playfield, height - i + 1, garbage)
        end

    elseif mode == "random" then
        for i = 1, lines do
            local hole = math.random(1, width)
            local garbage = excitingBrickGameMovement.generateGarbageRow(width, hole)

            excitingBrickGameMovement.setRow(playfield, height - i + 1, garbage)
        end
    end
end

-- Clear filled lines and move lines above down
-- Returns the cleared lines
function excitingBrickGameMovement.clearLines(playfield, rows)
    rows = rows or excitingBrickGameMovement.findFilledLines(playfield)

    for _, row in ipairs(rows) do
        excitingBrickGameMovement.moveFieldDown(playfield, row, 1)
    end

    return rows
end

-- See if the piece is allowed to move
-- Keep position if it collides, otherwise restore original
-- Returns whether the piece moved or not
function excitingBrickGameMovement.moveexcitingBrickShape(excitingBrickShape, playfield, offsetX, offsetY)
    local previousX, previousY = excitingBrickShape.x, excitingBrickShape.y

    excitingBrickShape.x = previousX + offsetX
    excitingBrickShape.y = previousY + offsetY

    local collided = excitingBrickGameMovement.exciting_brick_shapeColliding(excitingBrickShape, playfield)

    if collided then
        excitingBrickShape.x = previousX
        excitingBrickShape.y = previousY
    end

    return not collided
end

-- Check if the excitingBrickShape can move
function excitingBrickGameMovement.canMoveexcitingBrickShape(excitingBrickShape, playfield, offsetX, offsetY)
    local previousX, previousY = excitingBrickShape.x, excitingBrickShape.y

    excitingBrickShape.x = previousX + offsetX
    excitingBrickShape.y = previousY + offsetY

    local collided = excitingBrickGameMovement.exciting_brick_shapeColliding(excitingBrickShape, playfield)

    excitingBrickShape.x = previousX
    excitingBrickShape.y = previousY

    return not collided
end

-- Rotate excitingBrickShape
-- Returns whether the rotated or not
function excitingBrickGameMovement.rotateexcitingBrickShape(excitingBrickShape, playfield, direction)
    local offsets = excitingBrickShape.offsets
    local testCount = #offsets[1]

    local initialX, initialY = excitingBrickShape.x, excitingBrickShape.y
    local initialRotation = excitingBrickShape.rotation

    excitingBrickShape.rotation = utils.mod1(excitingBrickShape.rotation + direction, #excitingBrickShape.rotations)

    local newRotation = excitingBrickShape.rotation

    local currentTests = offsets[initialRotation]
    local newTests = offsets[newRotation]

    for test = 1, testCount do
        local cx, cy = currentTests[test][1], currentTests[test][2]
        local nx, ny = newTests[test][1], newTests[test][2]
        local ox, oy = cx - nx, cy - ny

        -- Our Y axis is flipped
        excitingBrickShape.x = initialX + ox
        excitingBrickShape.y = initialY - oy

        local collided = excitingBrickGameMovement.exciting_brick_shapeColliding(excitingBrickShape, playfield)

        if not collided then
            return true, ox, -oy, test
        end
    end

    excitingBrickShape.x = initialX
    excitingBrickShape.y = initialY
    excitingBrickShape.rotation = initialRotation

    return false
end

return excitingBrickGameMovement
