local ui = require("ui")
local uiElements = require("ui.elements")
local uiUtils = require("ui.utils")

local fonts = require("fonts")
local excitingBrickGameSceneHandler = require("mods").requireFromPlugin("exciting_brick_game.exciting_brick_game_scene_handler")
local excitingBrickGameUtils = require("mods").requireFromPlugin("utils")
local menubar = require("ui.menubar")
local widgetUtils = require("ui.widgets.utils")

local windowPersister = require("ui.window_position_persister")
local windowPersisterName = "exciting_brick_game_window"

local gameWindow
local gameWindowGroup = uiElements.group({})

local mainWindow = {}

if not ui._addedJoystickHooks then
    local origGamepadPressed = love.gamepadpressed
    local origGamepadReleased = love.gamepadrelease

    love.gamepadpressed = function(...)
        if origGamepadPressed then
            origGamepadPressed(...)
        end

        ui.interactiveIterate(ui.focusing, "gamepadpressed", ...)
    end

    love.gamepadreleased = function(...)
        if origGamepadReleased then
            origGamepadReleased(...)
        end

        ui.interactiveIterate(ui.focusing, "gamepadreleased", ...)
    end

    ui._addedJoystickHooks = true
end

local function gameHandlerProxy(handler, functionName)
    return function(self, ...)
        if handler[functionName] then
            return handler[functionName](handler, ...)
        end
    end
end

local function gamehandlerUpdate(handler)
    return function(widget, dt)
        widget:repaint()

        if handler._gameHandler and (not ui.root.focused or ui.focusing ~= widget) and not handler._gameHandler:anyGamesPaused() then
            handler._gameHandler:pauseAllGames()
        end

        handler:update(dt)
    end
end

local modInfo, pluginInfo = require("mods").findLoadedMod("ProcrastinationHelper")
local bgImage = love.graphics.newImage(string.format("%s/themes/default/game_window.png", pluginInfo._mountPointLoenn))

local function gamehandlerDraw(handler)
    return function(widget)
        local previousFont = love.graphics.getFont()
        local origWindowDimensions = love.graphics.getDimensions

        love.graphics.getDimensions = function()
            return widget.width, widget.height
        end

        love.graphics.push()
        love.graphics.setFont(fonts.font)
        love.graphics.translate(widget.screenX, widget.screenY)

        love.graphics.draw(bgImage, 0, 0)
        handler:draw()

        if not ui.root.focused or ui.focusing ~= widget then
            handler._theme:drawUnfocusedMessage()
        end

        love.graphics.setFont(previousFont)
        love.graphics.pop()

        love.graphics.getDimensions = origWindowDimensions
    end
end

function mainWindow.createWindow()
    math.randomseed(os.time())

    local gameHandler = excitingBrickGameSceneHandler.createScene()
    local gameWidget = uiElements.new({
        width = 740,
        height = 700,

        interactive = 1,

        onKeyPress = gameHandlerProxy(gameHandler, "keypressed"),
        onKeyRelease = gameHandlerProxy(gameHandler, "keyreleased"),
        gamepadpressed = gameHandlerProxy(gameHandler, "gamepadpressed"),
        gamepadreleased = gameHandlerProxy(gameHandler, "gamepadreleased"),
        update = gamehandlerUpdate(gameHandler),
        draw = gamehandlerDraw(gameHandler)
    })

    gameWindow = uiElements.window("Exciting Brick Game!", gameWidget)

    gameHandler._mainMenu._options[4].callback = function()
        gameWindow:removeSelf()
    end

    gameHandler._theme._fontScale = 2

    windowPersister.trackWindow(windowPersisterName, gameWindow)
    --widgetUtils.addWindowCloseButton(gameWindow)
    gameWindowGroup.parent:addChild(gameWindow)

    ui.runLate(function ()
        gameWidget.focused = true
        ui.focusing = gameWidget
    end)
end

function mainWindow.subMenuCallback()
    mainWindow.createWindow()
end

function mainWindow.getWindow()
    local addedMenubar = false

    for _, item in pairs(menubar.menubar) do
        if type(item) == "table" then
            if item[1] == "procrastinate" then
                addedMenubar = true
            end
        end
    end

    if not addedMenubar then
        table.insert(menubar.menubar, {"procrastinate", mainWindow.subMenuCallback})
    end

    return gameWindowGroup
end

return mainWindow