local matrixLib = require("mods").requireFromPlugin("matrix")

local utils = {}

-- Based on Selene implementation
function utils.split(text, sep, maxSize, useRegex)
    local res = {}

    if not sep or sep == "" then
        sep = 1
    end

    if type(sep) == "number" then
        while #text > 0 do
            table.insert(res, text:sub(1, sep))
            text = text:sub(sep + 1)
        end

    else
        -- http://lua-users.org/wiki/SplitJoin
        -- "Function: true Python semantics for split"
        assert(maxSize == nil or maxSize >= 1, "invalid argument 3: max table size is smaller than 1.")

         if #text > 0 then
            maxSize = maxSize or -1

            local plain = not useRegex

            local tblIndex, searchIndex = 1, 1
            local start, stop = text:find(sep, searchIndex, plain)

            while start and maxSize ~= 0 do
                res[tblIndex] = text:sub(searchIndex, start - 1)
                tblIndex = tblIndex + 1
                searchIndex = stop + 1
                start, stop = text:find(sep, searchIndex, plain)
                maxSize = maxSize - 1
            end

            res[tblIndex] = text:sub(searchIndex)
        end
    end

    return res
end

function utils.isInteger(n)
    return n == math.floor(n)
end

function utils.tileImageToQuadMatrix(image, tileWidth, tileHeight)
    local imageWidth, imageHeight = image:getDimensions()
    local tilesHorizontal, tilesVertical = imageWidth / tileWidth, imageHeight / tileHeight

    if utils.isInteger(tilesHorizontal) and utils.isInteger(tilesVertical) then
        local matrix = matrixLib.filled(nil, tilesHorizontal, tilesVertical)

        for x = 1, tilesHorizontal do
            for y = 1, tilesVertical do
                local quad = love.graphics.newQuad((x - 1) * tileWidth, (y - 1) * tileHeight, tileWidth, tileHeight, imageWidth, imageHeight)

                matrix:set(x, y, quad)
            end
        end

        return matrix

    else
        -- TODO - Warn
    end
end

-- TODO - Verify quadMatrix size
-- TODO - Support quadMatrix larger than 3x3
function utils.drawNinePatch(image, quadMatrix, drawX, drawY, tilesHorizontal, tilesVertical, tileWidth, tileHeight, drawMiddle, color)
    local pr, pg, pb, pa

    if color then
        pr, pg, pb, pa = love.graphics.getColor()

        love.graphics.setColor(color)
    end


    for x = 1, tilesHorizontal do
        for y = 1, tilesVertical do
            local quadX = x == 1 and 1 or (x == tilesHorizontal and 3 or 2)
            local quadY = y == 1 and 1 or (y == tilesVertical and 3 or 2)

            if not drawMiddle and quadX == 2 and quadY == 2 then

            else
                local quad = quadMatrix:get(quadX, quadY)

                love.graphics.draw(image, quad, drawX + (x - 1) * tileWidth, drawY + (y - 1) * tileHeight)
            end
        end
    end

    if color then
        love.graphics.setColor(pr, pg, pb, pa)
    end
end

function utils.getTextSize(text, font, fontSize)
    font = font or love.graphics.getFont()
    fontSize = fontSize or 1

    local width = font:getWidth(text) * fontSize
    local height = font:getHeight(text) * fontSize

    return width, height
end

function utils.printCenteredText(text, x, y, width, height, font, fontSize, textColor)
    font = font or love.graphics.getFont()
    fontSize = fontSize or 1

    local pr, pg, pb, pa

    if textColor then
        pr, pg, pb, pa = love.graphics.getColor()

        love.graphics.setColor(textColor)
    end

    local fontHeight = font:getHeight()
    local fontLineHeight = font:getLineHeight()
    local longest, lines = font:getWrap(text, width / fontSize)
    local textHeight = (#lines - 1) * (fontHeight * fontLineHeight) + fontHeight

    local offsetX = 1
    local offsetY = math.floor((height - textHeight * fontSize) / 2) + 1

    love.graphics.push()

    love.graphics.translate(x + offsetX, y + offsetY)
    love.graphics.scale(fontSize, fontSize)

    love.graphics.printf(text, 0, 0, width / fontSize, "center")

    love.graphics.pop()

    if textColor then
        love.graphics.setColor(pr, pg, pb, pa)
    end
end

function utils.deepcopy(v)
    if type(v) == "table" then
        local res = {}

        for key, value in pairs(v) do
            res[key] = utils.deepcopy(value)
        end

        return res

    else
        return v
    end
end

function utils.mod1(n, d)
    local m = n % d

    return m == 0 and d or m
end

return utils